/********************************************************

	Classes for reading & writing AVIs
	Copyright 2000 Eugene Smith (divx@euro.ru)
	Last modified: 03.08.2000

*********************************************************/
#ifndef _AVIFILE_H
#define _AVIFILE_H

#include <avifmt.h>
#include <default.h>
#include <wine/vfw.h>
#include <except.h>
class AviStream
{
protected:
    AVIStreamHeader m_header;
public:
    enum StreamType{
    Audio,
    Video,
    Other
    };
};

/**
 *
 *   
 *    This class provides means of sequential access to data from one 
 * stream, usually video or soundtrack. Though AVI files rarely contain
 * more than one soundtrack or streams of other kind, they are also supported.
 *
 *   It maintains an internal position counter, which can be read with 
 * @ref GetPos()/@ref GetTime() and changed with @ref Seek()/@ref SeekToTime().
 * This counter is modified by all reading functions; function that reads
 * 10 frames will increase counter by 10.
 *
 *   IAviReadStream is capable of implicit video and audio decompression,
 * so that you won't need to work with decoders directly. All you need is
 * to call @ref StartStreaming(), and if it succeeds ( return value is zero ),
 * call members that decompress data into internal buffer ( @ref ReadFrame() ),
 * return this buffer ( @ref GetFrame() ) or decompress data into your memory
 * ( @ref ReadFrames() ).
 *
 *   If you want to have direct access to compressed data, use @ref ReadDirect().
 * Note that this call is incompatible with streaming, so calls to @ref ReadDirect()
 * between @ref StartStreaming() and @ref StopStreaming() will fail. It is done so
 * because dropping even one data frame during sequence decompression will ruin 
 * current picture.
 * 
 *
 */
class IAviReadStream : public AviStream
{
public:
    virtual ~IAviReadStream();
    virtual StreamType GetType() const					=0;
    /**
     * For video stream returns video format information ( usually in
     * BITMAPINFOHEADER format ). For other stream types fails.
     */
    virtual HRESULT GetVideoFormatInfo(void* bi)			=0;
    /**
     * For audio stream returns audio format information. Stores main
     * format info in   bi  - WAVEFORMATEX structure - and pointer
     * to extended compressor-specific data in   ext  if it's nonzero. 
     * If *ext!=NULL on return, you should free it with   delete 
     * when it's not needed anymore.
     *
     * For other stream types fails.
     */
    virtual HRESULT GetAudioFormatInfo(void* bi, char** ext) 		=0;
    
    /**
     * Total length of stream in samples.
     */
    virtual unsigned GetEndPos() const					=0;	
    /**
     * Total length of stream in seconds.
     */
    virtual double GetEndTime() const 					=0;
    /**
     * Duration of one frame/sample.
     */
    virtual double GetFrameTime() const 				=0;

    //these two are only meaningful for video streams, but no checking
    //is done
    virtual unsigned GetNextKeyFrame() const 				=0;
    virtual unsigned GetPrevKeyFrame() const				=0;
    
    virtual int ToNextKeyFrame() 					=0;
    virtual int ToPrevKeyFrame() 					=0;
    /**
     * Seeks to position   pos  in stream. In video streams, if
     * streaming is started, it will need to decompress all
     * video frames since last key-frame before pos, and thus can 
     * be quite slow. 
     */    
    virtual HRESULT Seek(unsigned pos) 					=0;
    /** 
     * For video stream seeks to nearest keyframe.
     */
    virtual double SeekToTime(double time) 				=0;
    /**
     * Current time in stream.
     */
    virtual double GetTime() const					=0;
    /**
     * Current position in stream.
     */
    virtual unsigned GetPos() const					=0;
    /**
     * Returns flags associated with current frame. You'll need this
     * function if you use VideoDecoder for decompression. Then,
     * pass resulting value to decoder together with frame data.
     */
    virtual HRESULT GetFrameFlags(int* flags) 				=0;

    // Initializes decoders.
    virtual HRESULT StartStreaming() 					=0;
    virtual int IsStreaming() 						=0;
    /**
     * Specifies desired output format. Call after StartStreaming.
     * Only valid for video streams.
     */
    virtual HRESULT SetOutputFormat() 					=0;
    /**
     * Returns current output format. Call after StartStreaming.
     */
    virtual HRESULT GetOutputFormat(void* format, unsigned size)	=0;
    virtual HRESULT SetBitDepth(unsigned depth) =0;
    /**
     * Reads one frame, decompresses it into internal buffer
     * and increases position counter.
     */
    virtual HRESULT ReadFrame() 					=0;
    /**
     * Returns pointer to internal frame buffer.
     */
    virtual char* GetFrame() 						=0;		
    virtual int GetFrameSize() 						=0;
    /**
     * Reads and decompresses variable number of frames into
     * user-supplied buffer.
     */
    virtual HRESULT ReadFrames(char* buffer, unsigned bufsize,
	 unsigned samples,
	 unsigned& samples_read, unsigned& bytes_read)			=0;
    /**
     * Closes decoders.     
     */
    virtual HRESULT StopStreaming() 					=0; 

    /**
     * Directly reads data from file. Incompatible with
     * streaming.
     */
    virtual HRESULT ReadDirect(char* buffer, unsigned bufsize,
         unsigned samples,
	 unsigned& samples_read, unsigned& bytes_read) 			=0;
    /**
     * Checks for end of stream. Nonzero if true.
     */
    virtual HRESULT Eof() const =0;
};
    

class IAviReadFile
{
public:
    virtual ~IAviReadFile();
    virtual unsigned StreamCount() 					=0;
    virtual unsigned VideoStreamCount() 				=0;
    virtual unsigned AudioStreamCount() 				=0;
    virtual IAviReadStream* GetStream(unsigned stream_id,
	 AviStream::StreamType type) 					=0;
    virtual HRESULT GetFileHeader(MainAVIHeader* header) 		=0;
//    virtual int AddRef() =0;
//    virtual int Release() =0;
};
    
class IAviWriteStream : public AviStream
{
friend class IAviWriteFile;
public:
    virtual ~IAviWriteStream() 						=0;
    //
    // these two should be called before data insertion begins
    //
    //
    // for video streams, specify time in microsecs per frame in frame_rate
    // for audio streams - count of bytes per second
    //   
    virtual void SetHeader(int handler, int frame_rate,
	int samplesize=0, int quality=0, int flags=0) 			=0;
    virtual void SetFormat(const char* format, unsigned format_size)  	=0;
    
    virtual HRESULT AddChunk(const char* chunk, 
	    unsigned size, unsigned flags=0) 				=0;

    virtual unsigned GetLength() 					=0;
    virtual StreamType GetType() const 					=0;
};            

class IAviVideoWriteStream : public AviStream
{
public:
    virtual ~IAviVideoWriteStream();
    virtual void SetHeader(int frame_rate,
            int flags=0) 						=0;
    virtual HRESULT Start()						=0;
    virtual HRESULT AddFrame( char* chunk)			 	=0;
    virtual HRESULT Stop()						=0;
    virtual HRESULT SetQuality(int quality)				=0;
    virtual HRESULT SetKeyFrame(int frequency)				=0;
    virtual unsigned GetLength() 					=0;
};

class IAviWriteFile
{
friend class IAviWriteStream;
public:
    virtual ~IAviWriteFile();
    virtual IAviWriteStream* AddStream(enum AviStream::StreamType) 	=0;
    virtual IAviVideoWriteStream* AddVideoStream(int fourcc, 
				    BITMAPINFOHEADER* srchdr)		=0;
    virtual HRESULT Reserve(unsigned size)				=0;
    virtual HRESULT WriteChunk(int fourcc, char* data, int size)	=0;
};

IAviReadFile* CreateIAviReadFile(const char* name);    
IAviWriteFile* CreateIAviWriteFile(const char* name, int flags=0, int mask=00777);
    
#endif