/********************************************************

	Classes for writing AVIs
	Copyright 2000 Eugene Smith (divx@euro.ru)
	Last modified: 03.08.2000

*********************************************************/


#include <avifmt.h>
#include <default.h>
#include <unistd.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <pthread.h>
#include <avifile.h>    
#include <videoencoder.h>

#include <vector>
using namespace std;

#include <except.h>
#define __MODULE__ "AviWriteFile"

class AviWriteFile;

class AviWriteStream : public IAviWriteStream
{
friend class AviWriteFile;

protected:
    AviWriteFile* m_file;
    int m_status;
    int m_fd;
    enum StreamState
    {
	ValidFormat = 1,
	ValidHeader = 2,
	ValidType = 4,
	ValidStream = 7
    };
    StreamType m_type;
    
    char* m_format;
    unsigned m_forsize;
    int m_ckid;
    void SetType(StreamType type)
    {
	if(type==Other)
	    throw FATAL("Unsupported stream type");
	m_status|=ValidType;
	m_type=type;
	return;
    }	 
public:
    AviWriteStream(AviWriteFile* file, int ckid);
    
    ~AviWriteStream();
    //
    // these two should be called before data insertion begins
    //
    //
    // for video streams, specify time in microsecs per frame in frame_rate
    // for audio streams - count of bytes per second
    //   
    void SetHeader(int handler, int frame_rate, int samplesize=0, int quality=0, int flags=0)
    {
	if(!(m_status & ValidType))
	    throw FATAL("Specify type before header");
	
	m_status|=ValidHeader;
	memset(&m_header, 0, sizeof(m_header));
	m_header.fccType=((m_type==Video)?streamtypeVIDEO:streamtypeAUDIO);
	m_header.fccHandler=handler;
	m_header.dwFlags=flags;
	if(m_type==Video)
	{
	    m_header.dwRate=1000000;
	    m_header.dwScale=frame_rate;
	}
	else
	{
	    m_header.dwRate=frame_rate;
	    m_header.dwScale=samplesize;
	}        
	m_header.dwSampleSize=samplesize;
	m_header.dwQuality=quality;
	m_header.dwFlags=flags;
	return;
    }	
    void SetFormat(const char* format, unsigned format_size)
    {
	if(format==0)
	    throw FATAL("Bad format");
	m_status|=ValidFormat;
	m_format=new char[format_size];
	m_forsize=format_size;
	memcpy(m_format, format, format_size);
	if(m_type==Video)
	{
	    m_header.rcFrame.right=*(int*)(format+4);
	    m_header.rcFrame.bottom=*(int*)(format+8);
	}    
	return;
    }	
    StreamType GetType() const
    {
    	return m_type;
    }	    

    HRESULT AddChunk(const char* chunk, unsigned size, unsigned flags=0);

    unsigned GetLength()//0 on error
    {
	if(m_status!=ValidStream)
	    throw FATAL("Not a valid stream");
	return m_header.dwLength;
    }	    
};            


class AviVideoWriteStream : public IAviVideoWriteStream, public AviWriteStream
{
    VideoEncoder m_encoder;
    int m_vstatus;
    char* m_local_buffer;
    int m_fourcc;
public:
    AviVideoWriteStream(AviWriteFile* file, int ckid):AviWriteStream(file,ckid), m_vstatus(0),
	m_local_buffer(0){}
    ~AviVideoWriteStream(){m_encoder.Close();}
    void Create(int fourcc, BITMAPINFOHEADER* srchdr)
    {
	m_encoder.Init(fourcc, (const char*)srchdr);
	m_vstatus=1;
	m_fourcc=fourcc;
	SetType(Video);
        return;
    }   
    virtual void SetHeader(int frame_rate,
            int flags=0)
    {
	if(m_vstatus!=1)throw FATAL("Bad status");
	AviWriteStream::SetHeader(m_fourcc, frame_rate, 0, m_encoder.GetQuality(), flags);
    }
    virtual HRESULT Start()
    {
	if(m_vstatus!=1)throw FATAL("Bad status");
	BITMAPINFOHEADER bh=m_encoder.QueryOutputFormat();
	AviWriteStream::SetFormat((const char*)&bh, sizeof bh);
	m_encoder.Start();
	m_local_buffer=new char[m_encoder.QueryOutputSize()];
	if(m_local_buffer==0)throw MEMORY_FATAL;
	m_vstatus=2;
	return 0;
    }	
    virtual HRESULT Stop()
    {
	m_encoder.Stop();
	return 0;
    }	
    HRESULT AddFrame(char* chunk)
    {
	if(m_vstatus!=2)return -1;
	int size;
	int is_keyframe;
	HRESULT hr;
	if(chunk)
	{
	    hr=m_encoder.EncodeFrame(chunk, m_local_buffer, &is_keyframe, &size);
	    if(hr==0)
		hr=AviWriteStream::AddChunk(m_local_buffer, size, is_keyframe);
	}
	else hr=AviWriteStream::AddChunk(NULL, 0);
	return hr; 
    }
    virtual HRESULT SetQuality(int quality)
    {
	return m_encoder.SetQuality(quality);
    }	
    HRESULT SetKeyFrame(int frequency)
    {
	return m_encoder.SetKeyFrame(frequency);
    }	
    unsigned GetLength()
    {
	return AviWriteStream::GetLength();
    }	
};

//
//maintains global file info
//
class AviWriteFile: public IAviWriteFile
{
friend class AviWriteStream;
private:
    vector<AviWriteStream*> m_streams;
    int m_strcnt;
    MainAVIHeader m_header;
    int m_status;
    vector<AVIINDEXENTRY> m_index;
protected:
    int m_fd;
    void _destruct();
public:
    AviWriteFile(const char* name, int flags=0, int mask=00777);
    ~AviWriteFile();
//    void Create(const char* name, int flags=0, int mask=00777);
    IAviVideoWriteStream* AddVideoStream(int fourcc, 
				    BITMAPINFOHEADER* srchdr);
    IAviWriteStream* AddStream(enum AviStream::StreamType);

    HRESULT Reserve(unsigned size);
    HRESULT WriteChunk(int fourcc, char* data, int size);

protected:
    //
    //called when data is added to stream
    //
    void AddChunk(offset_t offset, unsigned size, unsigned id, unsigned flags=0);    
};
