/********************************************************

	Audio renderer object
	Copyright 2000 Eugene Smith (divx@euro.ru)
	Last modified: 04.08.2000

*********************************************************/

#define DEBUG_TOCONSOLE

#ifndef QUIET
#define debug_out(X) Debug cout<<X<<": "<<(t2-t1)/freq<<" ms"<<endl
#ifndef TIMING
#define Debug  if(debug_status==0)
#else
#define Debug 
#endif /*TIMING*/
#else
#define debug_out(X)
#define Debug if(0)
#endif /*QUIET*/

#include <stdio.h>
#include <iostream.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>


#ifdef USE_ESD
#include <esd.h>
#else
#include <sys/soundcard.h>
#endif

#include "aviplay_impl.h"
#include "audio.h"

#include <except.h>

#define __MODULE__ "AudioRenderer"
extern CPU_Freq freq;
static long long longcount()
{
    unsigned int i;
    unsigned int j;
    __asm__ __volatile__("rdtsc\n\t":"=a"(i), "=d"(j):);
    return ((long long)j<<32)+(long long)i;
}

static inline float to_float(long long t2, long long t1)
{
    long long qw=t2-t1;
    qw=(long long)(qw/freq);//milliseconds;
    return ((int)qw)/1000.;
}
template<class T>
inline T min(const T x, const T y){if(x<y)return x;else return y;}
template<class T>
inline T max(const T x, const T y){if(x>y)return x;else return y;}


audio_queue::audio_queue()
{
    pthread_mutex_init(&mutex, NULL);
    pthread_cond_init(&cond, NULL);
    frame_in=frame_out=frame_size=0;
}
audio_queue::~audio_queue()
{
    pthread_cond_destroy(&cond);
    pthread_mutex_destroy(&mutex);
}

int audio_queue::push(const char* data, int size)
{
    pthread_mutex_lock(&mutex);
    {
	int new_pos = frame_in+size;
	if(frame_in+size>sizeof(audio_frame))size=sizeof(audio_frame)-frame_in;
	memcpy(audio_frame+frame_in, data, size);
	while(new_pos > frame_max) 
        {
	    memcpy(audio_frame, audio_frame + frame_max, new_pos - frame_max);
	    new_pos  = new_pos - frame_max;
	}	
        frame_in = new_pos;
	frame_size += size;
    }
    pthread_mutex_unlock(&mutex);
    broadcast();
}
int audio_queue::write(int fd)//writes some data
{
    int startpos, size, tmp;
#ifdef __FreeBSD__
    const int AUDIO_BUFFER_SIZE=2048;
#else
    const int AUDIO_BUFFER_SIZE=8192;
#endif
    pthread_mutex_lock(&mutex);
    {
        size = AUDIO_BUFFER_SIZE;
	if((frame_in > frame_out) && (frame_in - frame_out < AUDIO_BUFFER_SIZE)) size = frame_in - frame_out;
        if((frame_in < frame_out) && (frame_max - frame_out < AUDIO_BUFFER_SIZE)) size = frame_max - frame_out;
    
	tmp=frame_in-frame_out;
        while(tmp<0)tmp+=frame_max;
	if(tmp!=frame_size)
        {
    	    printf("OOPS: frame_size!=tmp(%d,%d,%d)\n", frame_out, frame_in, frame_size);
        }    
	startpos=frame_out;
        if(frame_out == frame_max) frame_out = 0;
    }
    pthread_mutex_unlock(&mutex);
    //it'd be convenient to write() while mutex is locked, but
    //writes to ESD socket are very slow, mutex would be locked
    //all the time, and we need to gain access to it to write
    //data in AudioRenderer::doAudioExtract().
    int answer=::write(fd, audio_frame+startpos, size);
    if(answer==0)return 0;//sound card doesn't accept any more data ( e.g. DMA buffer is full )
    pthread_mutex_lock(&mutex);
    {
	if(frame_size==0)
	    //someone has cleared buffers while we were writing
	    //forget what we wrote, user should reset device
	{
	    pthread_mutex_unlock(&mutex);
	    return -1;
	}    
        frame_out += answer;
	if(frame_size<answer)
	    printf("OOPS: wrote %d bytes, buffer size %d\n", answer, frame_size);
	frame_size -= answer;
    }
    pthread_mutex_unlock(&mutex);
    return answer;
}
void audio_queue::clear()
{
    pthread_mutex_lock(&mutex);
    {
	frame_in=frame_out=frame_size=0;
    }	
    pthread_mutex_unlock(&mutex);
    return;
}

//thread func
void* doAudioOut(void* arg)
{
    AudioRenderer& a=*(AudioRenderer*)arg;

    int Snd_Limit=a.Snd_Limit;
    while(1)
    {
	if(a.quit)
	{
	    cout<<"Exiting audio thread"<<endl;
	    return 0;
	}    
	if(a.hangup)
	{
	    a.initialized=0;
	    a.hangup=0;
	}
	if(a.paused)
	{
	    usleep(100000);
//	    cout<<"paused"<<endl;
	    continue;
	}    
	static int reset_dev=0;
	if(!a.initialized)
	{	
	    if(reset_dev)
	    {
		a.queue.clear();
#ifndef USE_ESD
		ioctl(a.audio_fd, SNDCTL_DSP_RESET, 0);
#endif
		reset_dev=0;
	    }
	    usleep(100000);
//	    cout<<"Not initialized"<<endl;
	    continue;	    	    
	}else reset_dev=1;	    
	if(a.queue.wait()!=0)
	{
	    //wait() was interrupted, maybe by signal or quit=1
//	    cout<<"Bad wait"<<endl;
	    continue;
	}    
#ifndef USE_ESD
	audio_buf_info zz;
	ioctl(a.audio_fd, SNDCTL_DSP_GETOSPACE, &zz);
	if(zz.bytes==0)
	{
	    usleep(10000);
//	    cout<<"Buffer full"<<endl;
	    continue;
	}    
#endif
//	pthread_mutex_lock(&a.mutex);
/*	cout<<">"<<a.audiostream->GetTime()<<" "<<a.buffer_time()
		<<" "<<" -> "<<a.audiostream->GetTime()-a.buffer_time()
	        <<" "<<longcount()/550000000.<<endl;*/
	float in=a.getTime();
	long long ts=longcount();
//	cout<<"Before: "<<in<<endl;
//	pthread_mutex_unlock(&a.mutex);		
	int result=a.queue.write(a.audio_fd);
//	cout<<"Wrote "<<result<<" bytes"<<endl;
	if(result==0)
	    usleep(10000);
	else
	if(result<0)
	    continue;
	else
	{
	    pthread_mutex_lock(&a.mutex);
    	    a.audio_time=longcount();
	    double frame_time=a.buffer_time();
    	    a.audio_realpos=a.audiostream->GetTime()-frame_time;
//	    cout<<a.audiostream->GetTime()<<" "<<frame_time
//		<<" "<<" -> "<<a.audio_realpos
//	        <<" "<<longcount()/550000000.<<endl;
	    pthread_mutex_unlock(&a.mutex);
	    float out=a.getTime();
	    long long tx=longcount();
//	    cout<<"Write mismatch: "<<1000.*(out-in)-to_float(tx,ts)
//	    <<" ms ( writing "<<to_float(tx,ts)<<" ms, stream difference "<<1000.*(out-in)<<" ms )"<<endl;
	}    
    }
    return 0;
}

void AudioRenderer::doAudioExtract(double video_time)
{
    int t1, t2;
    int frames_written=0;
    
    const int one_frame_sound=10000;
    //This is the maximum amount of decompressed audio decoded 
    //each call to doAudioExtract(). By default it's called each 
    //frame, so we'll get up to 250 kb of audio data each frame
    //for 25 fps video. That should be enough for everybody, but 
    //even of it isn't, it shouldn't produce audio tearing.

    const int sound_size_max=176400;
    //Upper limit on size of audio queue. Large enough to prevent
    //tearing in case of big delays between doAudioExtract calls,
    //but very large values are unreasonable ( decompression takes
    // CPU time ). Current value is 1 sec of 44100/16/stereo.

    if(quit)
	return;

    if(queue.size()>sound_size_max)
        return;

    if(audiostream->Eof())
    {
	printf("Audio stream finished\n");
	initialized=0;
	queue.broadcast();
	return;
    }	
#ifdef USE_ESD    
    double actual_time=getTime();
    double async=video_time-actual_time;
    if((async<-.5)||(async>.5))
    {
#ifndef QUIET    
	cerr<<"WARNING: too large async"<<endl;
#endif	
	audiostream->SeekToTime(audiostream->GetTime()+async);
	return;
    }	
#endif
    
    unsigned ocnt;
    unsigned samples;
    char local_frame[one_frame_sound];
    pthread_mutex_lock(&mutex);
    audiostream->ReadFrames(local_frame, one_frame_sound,
         one_frame_sound, samples, ocnt);
    pthread_mutex_unlock(&mutex);
    t1=localcount();
    queue.push(local_frame, ocnt);
    return;
}

double AudioRenderer::getTime()
{
	pthread_mutex_lock(&mutex);
	if(time_start==0)
	{
	    audio_time=longcount();
	    time_start=longcount();
	    cout<<"Cleared audio_time"<<endl;
	}
	long long time_current=longcount();
	double actual_time=audio_realpos+to_float(time_current, audio_time);
//	cout<<audio_realpos<<"+"<<to_float(time_current, audio_time)<<"="<<actual_time<<" at "<<double(time_current-time_start)/550000000.-actual_time<<endl;
	pthread_mutex_unlock(&mutex);
        return actual_time;
}


AudioRenderer::AudioRenderer(IAviReadStream* as)
{
    try
    {
	audiostream=as;

        if(audiostream==0)
	    throw FATAL("NULL audiostream");
	int audio_status=audiostream->StartStreaming();
	if(audio_status!=0)
	    throw FATAL("Failed to start streaming");
	int tmp;
	audiostream->GetOutputFormat(&owf, sizeof owf);
#ifndef QUIET
	cout << "Audio format " << owf.nSamplesPerSec 
	<< "/" << owf.wBitsPerSample
	<< "/" << ((owf.nChannels!=1)?"stereo":"mono")
	<< endl;
#endif	
#ifndef USE_ESD
	audio_fd=-1;
	try
	{
    	    audio_fd=open("/dev/dsp",O_RDWR|O_NDELAY);
	    if(audio_fd<=0)
  		throw FATAL("Can't open audio device");
	    if(audio_fd>0)
	    {
  		int flag;
	        if((flag=fcntl(audio_fd,F_GETFL,0))<0)
  	    	    throw FATAL("fcntl");
		flag&=~O_NDELAY;
        	if(fcntl(audio_fd,F_SETFL,flag)<0)
		    throw FATAL("fcntl");
	    }
	}
	catch(FatalError&)
	{
	    if(audio_fd>0)::close(audio_fd);
	    audio_fd=open("/dev/dsp", O_RDWR);
	}
        if(audio_fd<=0)
     	    throw FATAL("Can't open audio device");
	ioctl(audio_fd, SNDCTL_DSP_RESET, NULL);
	audio_buf_info zz;
	ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &zz);
	Snd_Limit=zz.bytes;
	cout<<"Snd_Limit "<<Snd_Limit<<endl;
	tmp=owf.nChannels-1;
	if(ioctl(audio_fd, SNDCTL_DSP_STEREO, &tmp)!=0)
	    perror("ioctl(stereo)");
	tmp=owf.wBitsPerSample;
	if(ioctl(audio_fd, SNDCTL_DSP_SAMPLESIZE, &tmp)!=0)
	    perror("ioctl(samplesize)");
	tmp=owf.nSamplesPerSec;
	if(ioctl(audio_fd, SNDCTL_DSP_SPEED, &tmp)!=0)
	    perror("ioctl(speed)");
#else
	esd=esd_open_sound(NULL);
	if(esd<0)
	    throw FATAL("Can't open ESD connection");
	audio_fd=esd_play_stream(ESD_STREAM | ESD_PLAY | ((owf.nChannels!=1)?ESD_STEREO : ESD_MONO) |
	    ((owf.wBitsPerSample==16)?ESD_BITS16:ESD_BITS8),
	    owf.nSamplesPerSec, NULL, "aviplay");
	if(audio_fd==-1)
	    throw FATAL("Failed to open ESD stream");
#endif
	pthread_mutex_init(&mutex, 0);
    }
    catch(FatalError& error)
    {
	if(audio_fd>0)
	    ::close(audio_fd);
#ifdef USE_ESD		
	if(esd>0)
	    ::close(esd);	
#endif
	esd=audio_fd=-1;
	audiostream=0;
	throw;
    }	
    initialized=0;
    hangup=0;
    paused=0;
    audio_realpos=0;
    audio_time=longcount();
    time_start=0;	
    quit=0;
    audiostream->SeekToTime(0);
    pthread_create(&audio_thread, NULL, doAudioOut, (void*)this);
    return;
}

void AudioRenderer::start()
{
    if(initialized==1)
    {
	cout<<"Can't start(), already started"<<endl;
	return;
    }	
    audio_realpos=0;
    audio_time=longcount();
    quit=0;
    if(audiostream)
        audiostream->SeekToTime(0);
    queue.clear();
    time_start=0;
    paused=0;
    hangup=0;
    initialized=1;
}

void AudioRenderer::stop()
{
    initialized=0;
    queue.broadcast();
    time_start=0;
}
AudioRenderer::~AudioRenderer()
{
    cout<<"Destroying audio renderer"<<endl;
    quit=1;
    queue.broadcast();
    pthread_join(audio_thread, NULL);
    ::close(audio_fd);
#ifdef USE_ESD    
    if(esd>=0)
	esd_close(esd);
#endif    
    pthread_mutex_destroy(&mutex);
    cout<<"Destroy() successful"<<endl;
}
void AudioRenderer::reseek(double pos)
{
    if(audiostream)
        audiostream->SeekToTime(pos);
    if(initialized==0)return;
    hangup=1;
    wake();
    while(hangup)usleep(10000);
    queue.clear();
    char fake[16];
    memset(fake, 0, 16);
    queue.push(fake, 16);
#ifndef USE_ESD
    if(audio_fd!=0)
	ioctl(audio_fd, SNDCTL_DSP_RESET, NULL);
#endif	
    time_start=0;
    initialized=1;
    return;
}
double AudioRenderer::buffer_time()
{
    if(!initialized)return 0;
#ifndef USE_ESD
    audio_buf_info zz;
    ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &zz);
    
    double frame_time=double(queue.size()+(signed)Snd_Limit-zz.bytes)/owf.nChannels/2/owf.nSamplesPerSec;
#else
#warning FIXME: correct delay?
//    cout<<double(queue.size())/owf.nChannels/2/owf.nSamplesPerSec<<" # "<<double(latency(esd, owf))/owf.nChannels/2/owf.nSamplesPerSec <<endl;
    double frame_time=double(queue.size()+4*esd_get_latency(esd))/owf.nChannels/2/owf.nSamplesPerSec;
#endif
    return frame_time;
}
void AudioRenderer::clear()
{
    queue.clear();
}
void AudioRenderer::wake()
{
    queue.broadcast();
}